//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     cf.cxx
// Purpose:    Conditional Facts
// Last Change:06.04.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2001-2002 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include <stdlib.h>
#include "out.h"
#include "check.h"
#include "op.h"
#include "cf.h"

//=============================================================================
// Static Data Members (Class Attributes) of cf_c:
//=============================================================================

//-----------------------------------------------------------------------------
// Unique number of current comparison:
//-----------------------------------------------------------------------------

long cf_c::ComparID = 0;

//-----------------------------------------------------------------------------
// First conditional fact inserted in this hyperresolution iteration:
//-----------------------------------------------------------------------------

cf_c* cf_c::NewFirst = 0;

//-----------------------------------------------------------------------------
// Last conditional fact inserted in this hyperresolution iteration:
//-----------------------------------------------------------------------------

cf_c* cf_c::NewLast = 0;

//-----------------------------------------------------------------------------
// First conditional fact inserted in the previous iteration:
//-----------------------------------------------------------------------------

cf_c* cf_c::DiffFirst = 0;

//-----------------------------------------------------------------------------
// Last conditional fact inserted in the previous iteration:
//-----------------------------------------------------------------------------

cf_c* cf_c::DiffLast = 0;

//-----------------------------------------------------------------------------
// First conditional fact inserted before the previous iteration:
//-----------------------------------------------------------------------------

cf_c* cf_c::OldFirst = 0;

//-----------------------------------------------------------------------------
// Last conditional fact inserted before the previous iteration:
//-----------------------------------------------------------------------------

cf_c* cf_c::OldLast = 0;

//-----------------------------------------------------------------------------
// Conditional Facts that became obsolete due to a newly inserted one:
//-----------------------------------------------------------------------------

cf_c* cf_c::Obsolete = 0;

//-----------------------------------------------------------------------------
// The empty clause ("FALSE") was derived:
//-----------------------------------------------------------------------------

bool_t cf_c::Inconsistent = BOOL_FALSE;

//-----------------------------------------------------------------------------
// Unconditional disjunctions not yet processed for negative reduction:
//-----------------------------------------------------------------------------

cf_c* cf_c::NegRedQueue = 0;

//-----------------------------------------------------------------------------
// End of queue for negative reduction:
//-----------------------------------------------------------------------------

cf_c* cf_c::NegRedQueueLast = 0;

//=============================================================================
// Exported Static Member Functions (Class Methods) of cf_c:
//=============================================================================

//-----------------------------------------------------------------------------
// store: Store a conditional fact.
//-----------------------------------------------------------------------------

void cf_c::store(glit_t *head, int head_size, gneg_t *cond, int cond_size)
{
	int length;
	int i;
	cf_hd_c *head_elem;
	cf_cn_c *cond_elem;
	cf_c *cf;
	bool_t cond_answer;

	// Check parameters:
	CHECK(head != 0,
		"cf_c::store1: head is NIL");
	CHECK(head_size >= 0,
		"cf_c::store1: head_size is negative");
	CHECK(head_size <= VER_MAX_HEAD_LIT,
		"cf_c::store1: head_size too large");
	CHECK(cond != 0,
		"cf_c::store1: cond is NIL");
	CHECK(cond_size >= 0,
		"cf_c::store1: cond_size is negative");
	CHECK(cond_size <= VER_MAX_COND_NEG,
		"cf_c::store1: cond_size too large");

	// No further conditional facts can be stored after the empty clause:
	if(Inconsistent)
		return;

	// Get new comparison number:
	if(ComparID == VER_MAX_COMPAR) {
		out_c::too_many_compar(VER_MAX_COMPAR);
		exit(4);
	}
	ComparID++;

	// Check whether conditional fact is new (works only for length>0):
	length = head_size + cond_size;
	cond_answer = BOOL_FALSE;
	for(i = 0; i < head_size; i++) {
		CHECK(head[i]->valid(),
			"cf_c::store1: head literal is invalid");
		for(head_elem = head[i]->get_pos_occur(); head_elem != 0;
			head_elem = head_elem->get_next_fact_occur()) {
				if(head_elem->get_cf()->
						redundant(ComparID, length))
					return;
		}
		if(head[i]->is_answer_lit())
			cond_answer = BOOL_TRUE;
	}
	for(i = 0; i < cond_size; i++) {
		CHECK(cond[i]->valid(),
			"cf_c::store1: default negation is invalid");
		for(cond_elem = cond[i]->get_occur(); cond_elem != 0;
			cond_elem = cond_elem->get_next_neg_occur()) {
				if(cond_elem->get_cf()->
						redundant(ComparID, length))
					return;
		}
	}

	// If this is "FALSE", all conditional facts become obsolete:
	if(length == 0) {
		for(cf = NewFirst; cf != 0; cf = cf->get_next())
			cf->mark_obsolete();
		for(cf = DiffFirst; cf != 0; cf = cf->get_next())
			cf->mark_obsolete();
		for(cf = OldFirst; cf != 0; cf = cf->get_next())
			cf->mark_obsolete();
		Inconsistent = BOOL_TRUE;
	}

	// Create conditional fact:
	cf = new cf_c(cond_size == 0, cond_answer); // links it to the new list
	for(i = 0; i < head_size; i++) {
		head_elem = new cf_hd_c(head[i], cf);
		cf->add_head(head_elem);
	}
	for(i = 0; i < cond_size; i++) {
		cond_elem = new cf_cn_c(cond[i], cf);
		cf->add_cond(cond_elem);
	}
}

//-----------------------------------------------------------------------------
// store: Store a conditional fact.
//-----------------------------------------------------------------------------

void cf_c::store(lit_t *head_lits, int head_size,
			lit_t *neg_lits, int neg_size)
{
	glit_t head_facts[VER_MAX_HEAD_LIT];
	gneg_t cond_negs[VER_MAX_NEG_SIZE];
	int i;
	int num_head_facts;
	int num_cond_negs;
	glit_t fact;
	gneg_t neg;

	// Check parameters:
	CHECK(head_size >= 0,
		"cf_c::store2: head_size is negative");
	CHECK(head_size <= VER_MAX_HEAD_LIT,
		"cf_c::store2: head_size too large");
	CHECK(head_size == 0 || head_lits != 0,
		"cf_c::store2: head_lits is NIL");
	CHECK(neg_size >= 0,
		"cf_c::store2: neg_size is negative");
	CHECK(neg_size <= VER_MAX_NEG_SIZE,
		"cf_c::store2: neg_size too large");
	CHECK(neg_size == 0 || neg_lits != 0,
		"cf_c::store2: neg_lits is NIL");

	// Store head literals as facts:
	glit_c::dup_ck_begin();
	num_head_facts = 0;
	for(i = 0; i < head_size; i++) {
		CHECK(head_lits[i]->valid(),
				"cf_c::store2: invalid head literal");

		// Search fact or create new one:
		CHECK(i < VER_MAX_HEAD_LIT, "cf_c::store: i out of range");
		fact = glit_c::store(head_lits[i]);

		// Check for duplicate head literals:
		CHECK(num_head_facts <= i,
			"cf_c::store2: num_head_facts out of range");
		if(fact->dup_ck_is_new())
			head_facts[num_head_facts++] = fact;
	}
	glit_c::dup_ck_end();

	// Store groups of literals as default negations:
	num_cond_negs = 0;
	i = 0;
	gneg_c::dup_ck_begin();
	while(i < neg_size) {
		CHECK(neg_lits[i]->valid(),
				"cf_c::store2: invalid negated literal");

		// Search fact or create new one:
		neg = gneg_c::store(neg_lits + i);

		// Check for duplicate default negations:
		if(neg->dup_ck_is_new())
			cond_negs[num_cond_negs++] = neg;

		// Skip to next group of literals:
		while(i < neg_size && neg_lits[i] != 0)
			i++;
		if(i < neg_size)
			i++;
	}
	gneg_c::dup_ck_end();

	// Now store the conditional fact if not redundant:
	store(head_facts, num_head_facts, cond_negs, num_cond_negs);
}

//-----------------------------------------------------------------------------
// print_all: Print all currently existing conditional facts:
//-----------------------------------------------------------------------------

void cf_c::print_all(void)
{
	cf_c *cf;

	// Check for empty set:
	if(NewFirst == 0 && DiffFirst == 0 && OldFirst == 0) {
		out_c::no_cond_facts();
		return;
	}

	// Start unordered list (if HTML output):
	out_c::list_begin();

	// Loop over conditional facts generated before the previous iteration:
	for(cf = OldFirst; cf != 0; cf = cf->get_next()) {
		if(cf->is_obsolete() == BOOL_FALSE) {
			out_c::list_item();
			cf->print();
			out_c::print_nl();
		}
	}

	// Loop over conditional facts generated in the previous iteration:
	for(cf = DiffFirst; cf != 0; cf = cf->get_next()) {
		if(cf->is_obsolete() == BOOL_FALSE) {
			out_c::list_item();
			cf->print();
			out_c::print_nl();
		}
	}

	// Loop over conditional facts generated in this iteration:
	for(cf = NewFirst; cf != 0; cf = cf->get_next()) {
		if(cf->is_obsolete() == BOOL_FALSE) {
			out_c::list_item();
			cf->print();
			out_c::print_nl();
		}
	}

	// End unordered list (if HTML output):
	out_c::list_end();
}

//-----------------------------------------------------------------------------
// delete_obsolete: Delete the conditional facts that are now obsolete.
//-----------------------------------------------------------------------------

void cf_c::delete_obsolete(void)
{
	cf_c *cf;
	cf_c *next_cf;

	// Delete all conditional facts on the list of obsolete ones:
	for(cf = Obsolete; cf != 0; cf = next_cf) {
		next_cf = cf->get_next_obsolete();
		delete cf;
	}

	// Ok, there are currently no obsolete conditional facts:
	Obsolete = 0;
}

//-----------------------------------------------------------------------------
// next_iter: Start the next iteration of hyperresolution.
//-----------------------------------------------------------------------------

void cf_c::next_iter(void)
{
	// Append the Diff list to the Old list (update OldFirst/OldLast):
	if(OldLast == 0) {
		// Old list is empty, replace it by diff list:
		CHECK(OldFirst == 0, "cf_c::next_iter: Old list inconsistent");
		OldFirst = DiffFirst;
		OldLast = DiffLast;
	}
	else if (DiffFirst != 0) {
		// Both lists are non-empty, requires real append:
		CHECK(DiffLast !=0, "cf_c::next_iter: Diff list inconsistent");
		OldLast->set_next(DiffFirst);
		DiffFirst->set_prev(OldLast);
		OldLast = DiffLast;
	}
	else {
		// Nothing to be done if Diff list is empty
		CHECK(DiffLast == 0,
			"cf_c::next_iter: Diff list inconsistent 2");
	}

	// Replace Diff list by New list (update DiffFirst/DiffLast):
	DiffFirst = NewFirst;
	DiffLast = NewLast;

	// Initialize New list for next iteration (update NewFirst/NewLast):
	NewFirst = 0;
	NewLast = 0;
}

//-----------------------------------------------------------------------------
// reduce: Do positive and negative reduction.
//-----------------------------------------------------------------------------

void cf_c::reduce(void)
{
	cf_c *done;
	cf_c *next;
	glit_t pos_red_lit;

	// Repeat as long as positive or negative reduction is possible:
	for(;;) {
		cf_c::delete_obsolete();
		if(NegRedQueue != 0) {
			CHECK(NegRedQueue->valid(),
				"cf_c::reduce: NegRedQueue is invalid");
			CHECK(NegRedQueue->get_cond() == 0,
				"cf_c::reduce: NegRedQueue is inconsistent");

			// For each unconditional disjunctive fact,
			// do negative reduction:
			NegRedQueue->neg_reduce();

			// Remove from queue:
			done = NegRedQueue;
			next = done->get_neg_red_next();
			NegRedQueue = next;
			if(next != 0)
				next->set_neg_red_prev(0);
			else {
				CHECK(NegRedQueueLast == done,
					"cf_c::reduce: NegRedQueueLast wrong");
				NegRedQueueLast = 0;
			}
			done->set_neg_red_next(0);
			done->set_neg_red_prev(0);
		}
		else if((pos_red_lit = glit_c::fetch_pos_red_next()) != 0) {
			cf_c::pos_reduce(pos_red_lit);
		}
		else
			break;
	}

	// Move new facts (inserted via positive reduction) to diff/old state:
	next_iter();
}

//=============================================================================
// Exported Member Functions (Object Methods) of cf_c:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

cf_c::cf_c(bool_t unconditional, bool_t cond_answer)
{
	// Initialize data members:
	Next = 0;
	Prev = 0;
	Head = 0;
	LastHead = 0;
	Cond = 0;
	LastCond = 0;
	Length = 0;
	Overlap = 0;
	LastSet = 0;
	IsObsolete = BOOL_FALSE;
	IsCondAnswer = cond_answer;
	ObsoleteNext = 0;
	NegRedQueueNext = 0;
	NegRedQueuePrev = 0;
	CHECK_CODE(Magic = CF_MAGIC);

	// Add to list of new conditional facts:
	if(NewFirst == 0) {
		NewFirst = this;
		NewLast  = this;
	}
	else {
		CHECK(NewLast != 0, "cf_c::Constructor: NewLast not set");
		Prev = NewLast;
		NewLast->set_next(this);
		NewLast = this;
	}

	// If unconditional, add to the negative reduction queue:
	if(unconditional) {
		if(NegRedQueue == 0) {
			NegRedQueue = this;
			NegRedQueueLast = this;
		}
		else {
			CHECK(NegRedQueueLast != 0,
				"cf_c::Constructor: NegRedQueueLast not set");
			NegRedQueuePrev = NegRedQueueLast;
			NegRedQueueLast->set_neg_red_next(this);
			NegRedQueueLast = this;
		}
	}
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

cf_c::~cf_c(void)
{
	cf_hd_c *head_elem;
	cf_hd_c *next_head_elem;
	cf_cn_c *cond_elem;
	cf_cn_c *next_cond_elem;

	// When the destructor is called, the object must still be valid:
	CHECK(valid(), "cf_c::Destructor: Bad object");

	// Delete all head elements:
	for(head_elem = Head; head_elem != 0; head_elem = next_head_elem) {
		next_head_elem = head_elem->get_next_head_elem();
		delete head_elem;
	}

	// Delete all condition elements:
	for(cond_elem = Cond; cond_elem != 0; cond_elem = next_cond_elem) {
		next_cond_elem = cond_elem->get_next_cond_elem();
		delete cond_elem;
	}

	// Update forward pointer (first/next) to this conditional fact:
	if(NewFirst == this)
		NewFirst = Next;
	else if(DiffFirst == this)
		DiffFirst = Next;
	else if(OldFirst == this)
		OldFirst = Next;
	else {
		CHECK(Prev != 0,
			"cf_c::Destructor: No forward pointer to this cf");
		Prev->set_next(Next);
	}

	// Update backward pointer (last/prev) to this conditional fact:
	if(NewLast == this)
		NewLast = Prev;
	else if(DiffLast == this)
		DiffLast = Prev;
	else if(OldLast == this)
		OldLast = Prev;
	else {
		CHECK(Next != 0,
			"cf_c::Destructor: No backward pointer to this cf");
		Next->set_prev(Prev);
	}

	// Remove from negative reduction queue (list of disjunctive facts):
	if(Cond == 0) {
		if(NegRedQueueNext != 0)
			NegRedQueueNext->set_neg_red_prev(NegRedQueuePrev);
		if(NegRedQueuePrev != 0)
			NegRedQueuePrev->set_neg_red_next(NegRedQueueNext);
		if(NegRedQueue == this)
			NegRedQueue = NegRedQueueNext;
		if(NegRedQueueLast == this)
			NegRedQueueLast = NegRedQueuePrev;
	}

	// Invalidate the object:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
bool_t	cf_c::valid(void) const
{
	// Check object pointer:
	CHECK(this != 0, "cf_c::valid: NIL object");

	// Magic number must be correct:
	if(Magic != CF_MAGIC)
		return(BOOL_FALSE);

	// Next must be set if this is not the last conditional fact:
	CHECK(Next!= 0|| NewLast==this || DiffLast==this || OldLast==this,
		"cf_c::valid: Next not set");

	// Prev must be set if this is not the first conditional fact:
	CHECK(Prev!=0 || NewFirst==this || DiffFirst==this || OldFirst==this,
			"cf_c::valid: Prev not set");
	
	// Head and LastHead can only together be NIL or Not-NIL:
	CHECK(Head == 0 && LastHead == 0 || Head != 0 && LastHead != 0,
		"cf_c::valid: Head and LastHead can only together be NIL");

	// Cond and LastCond can only together be NIL or Not-NIL:
	CHECK(Cond == 0 && LastCond == 0 || Cond != 0 && LastCond != 0,
		"cf_c::valid: Cond and LastCond can only together be NIL");

	// The counters must all be not negative:
	CHECK(Length >= 0, "cf_c::valid: Length is negative");
	CHECK(Overlap >= 0, "cf_c::valid: Overlap is negative");
	CHECK(LastSet >= 0, "cf_c::valid: LastSet is negative");

	// IsObsolete must be a valid boolean value:
	CHECK(IsObsolete == BOOL_TRUE || IsObsolete == BOOL_FALSE,
		"cf_c::valid: IsObsolete is invalid");

	// ObsoleteNext can only be set if this conditional fact is obsolete:
	CHECK(IsObsolete == BOOL_TRUE || ObsoleteNext == 0,
		"cf_c::valid: ObsoleteNext is wrong");

	// Obsolete can only point to obsolete conditional facts:
	CHECK(IsObsolete == BOOL_TRUE || Obsolete != this,
		"cf_c::valid: Obsolete is wrong");

	// If this fact has a condition, it cannot be on the NegRedQueue:
	CHECK(Cond == 0 || NegRedQueueNext == 0,
		"cf_c::valid: NegRedQueueNext must be NIL for cond facts");
	CHECK(Cond == 0 || NegRedQueuePrev == 0,
		"cf_c::valid: NegRedQueuePrev must be NIL for cond facts");
	CHECK(Cond == 0 || NegRedQueue != this,
		"cf_c::valid: Cond facts cannot be on the NegRedQueue");
	CHECK(Cond == 0 || NegRedQueueLast != this,
		"cf_c::valid: Cond facts cannot be on the NegRedQueue [2]");

	// Ok:
	return(BOOL_TRUE);
}
#endif

//-----------------------------------------------------------------------------
// add_head: Add a head literal.
//-----------------------------------------------------------------------------

void cf_c::add_head(class cf_hd_c *head_elem)
{
	// Check consistency of this object:
	CHECK(valid(), "cf_c::add_head: Bad object");

	// Check parameter:
	CHECK(head_elem != 0, "cf_c::add_head: head_elem is NIL");
	CHECK(head_elem->valid(), "cf_c::add_head: head_elem is invalid");

	// Append head_elem to list of head literals:
	if(Head == 0)
		Head = head_elem;
	else
		LastHead->append(head_elem);
	LastHead = head_elem;

	// Increase length of clause:
	Length++;
}

//-----------------------------------------------------------------------------
// add_cond: Add a default negation (condition) to the body.
//-----------------------------------------------------------------------------

void cf_c::add_cond(class cf_cn_c *cond_elem)
{
	// Check consistency of this object:
	CHECK(valid(), "cf_c::add_cond: Bad object");

	// Check parameter:
	CHECK(cond_elem != 0, "cf_c::add_cond: cond_elem is NIL");
	CHECK(cond_elem->valid(), "cf_c::add_cond: cond_elem is invalid");

	// Append cond_elem to list of default negations (conditions):
	if(Cond == 0)
		Cond = cond_elem;
	else
		LastCond->append(cond_elem);
	LastCond = cond_elem;

	// Increase length of clause:
	Length++;
}

//-----------------------------------------------------------------------------
// redundant: Check for non-minimal conditional facts.
//-----------------------------------------------------------------------------

bool_t cf_c::redundant(long comparison, int length)
{
	// Check consistency of object:
	CHECK(valid(), "cf_c::redundant: Bad object");

	// Check parameters:
	CHECK(comparison >= 0, "cf_c::redundant: comparison is negative");
	CHECK(length > 0, "cf_c::redundant: length is not positive");

	// Was the Overlap counter already used in this comparison?
	if(LastSet == comparison)
		Overlap++;
	else {
		Overlap = 1;
		LastSet = comparison;
	}

	// If we reached the length of the old clause, the new is redundant:
	if(Overlap == Length)
		return(BOOL_TRUE);

	// If we reached the length of the new clause, the old is redundant:
	if(Overlap == length && IsObsolete == BOOL_FALSE) {
		IsObsolete = BOOL_TRUE;
		ObsoleteNext = Obsolete;
		Obsolete = this;
	}

	// Ok, the new clause is not (yet discovered as) redundant:
	return(BOOL_FALSE);
}

//-----------------------------------------------------------------------------
// print: Print this conditional fact:
//-----------------------------------------------------------------------------

void cf_c::print(void) const
{
	cf_hd_c *head_elem;
	cf_cn_c *cond_elem;
	str_t	or_sym;
	str_t	if_sym;
	str_t	and_sym;
	bool_t	first_time;

	// Check consistency of object:
	CHECK(valid(), "cf_c::print: Bad object");

	// Check for the contradiction:
	if(Length == 0) {
		out_c::print_str("FALSE.");
		out_c::print_nl();
		return;
	}

	// Get logical operator symbols:
	or_sym = op_c::op_or()->name();
	if_sym = op_c::op_if()->name();
	and_sym = op_c::op_and()->name();

	// Print head:
	first_time = BOOL_TRUE;
	for(head_elem = Head; head_elem != 0;
			head_elem = head_elem->get_next_head_elem()) {
		// Print "or" operator between facts (ground literals):
		if(first_time)
			first_time = BOOL_FALSE;
		else
			out_c::print_str(or_sym);

		// Print fact (ground literal):
		head_elem->get_fact()->print();
	}

	// Print condition:
	first_time = BOOL_TRUE;
	for(cond_elem = Cond; cond_elem != 0;
		cond_elem = cond_elem->get_next_cond_elem()) {
		// Print "if" before negations, "and" between them:
		if(first_time) {
			out_c::print_str(if_sym);
			first_time = BOOL_FALSE;
		}
		else
			out_c::print_str(and_sym);

		// Print condition (ground negation):
		cond_elem->get_neg()->print();
	}

	// Mark end of conditional fact:
	out_c::print_str(".");
}

//-----------------------------------------------------------------------------
// neg_reduce: Do one negative reduction step.
//-----------------------------------------------------------------------------

void cf_c::neg_reduce(void)
{
	cf_hd_t	dis_fact;
	gneg_t	negs[VER_MAX_HEAD_LIT];
	int	num_negs;
	gneg_l_t lit_occur;
	int i;
	cf_cn_t	cond_elem;

	// Check consistency of object:
	CHECK(valid(), "cf_c::neg_reduce: NegRedQueue is invalid");
	CHECK(Cond == 0, "cf_c::neg_reduce: NegRedQueue is inconsistent");

	// For each head literal, find the corresponding default negation:
	num_negs = 0;
	for(dis_fact = NegRedQueue->get_head(); dis_fact != 0;
			dis_fact = dis_fact->get_next_head_elem()) {
		// Check all negations that contain this ground literal:
		lit_occur = dis_fact->get_fact()->get_neg_occur();
		while(lit_occur != 0 && lit_occur->get_neg()->is_conjunctive())
			lit_occur = lit_occur->get_next_lit_occur();

		// Abort if no negation consists of only this literal:
		if(lit_occur == 0)
			return;

		// Otherwise remember the negation consisting of the
		// literal. Of course, there can be only one such neg.:
		CHECK(num_negs >= 0, "cf_c::neg_reduce: num_negs is invalid");
		CHECK(num_negs < VER_MAX_HEAD_LIT,
			"cf_c::reduce: num_negs is too large");
		negs[num_negs++] = lit_occur->get_neg();
	}

	// Get new comparison number:
	if(ComparID == VER_MAX_COMPAR) {
		out_c::too_many_compar(VER_MAX_COMPAR);
		exit(4);
	}
	ComparID++;

	// Eliminate conditional facts that contain all literals in negs:
	for(i = 0; i < num_negs; i++) {
		CHECK(negs[i]->valid(),
			"cf_c::neg_reduce: default negation is invalid");
		for(cond_elem = negs[i]->get_occur(); cond_elem != 0;
			cond_elem = cond_elem->get_next_neg_occur()) {
				cond_elem->get_cf()->neg_reduce_test(num_negs);
		}
	}
}

//-----------------------------------------------------------------------------
// neg_reduce_test: Check overlap of condition with disjunctive fact.
//-----------------------------------------------------------------------------

void cf_c::neg_reduce_test(int length)
{
	// Check consistency of object:
	CHECK(valid(), "cf_c::neg_reduce_test: Bad object");

	// Check parameter:
	CHECK(length > 0, "cf_c::neg_reduce_test: length is not positive");
	CHECK(length <= VER_MAX_HEAD_LIT,
		"cf_c::neg_reduce_test: length too large");

	// Was the Overlap counter already used in this comparison?
	if(LastSet == ComparID)
		Overlap++;
	else {
		Overlap = 1;
		LastSet = ComparID;
	}

	// If we reached the length of the disjucntion,
	// the condition can never be satisfied:
	if(Overlap == length && IsObsolete == BOOL_FALSE) {
		IsObsolete = BOOL_TRUE;
		ObsoleteNext = Obsolete;
		Obsolete = this;
	}
}

//-----------------------------------------------------------------------------
// pos_reduce: Do positive reduction.
//-----------------------------------------------------------------------------

void cf_c::pos_reduce(glit_t lit)
{
	gneg_l_t neg_elem;
	cf_cn_t  cond_elem;
	glit_t   new_head[VER_MAX_HEAD_LIT];
	glit_t   *head_ptr;
	gneg_t   new_cond[VER_MAX_COND_NEG];
	gneg_t   *cond_ptr;
	cf_t     cf;
	cf_hd_t  hd;
	cf_cn_t  cn;

	// Find negative literal that can be evaluated to "true":
	neg_elem = lit->get_neg_occur();
	while(neg_elem != 0 && neg_elem->get_neg()->is_conjunctive())
			neg_elem = neg_elem->get_next_lit_occur();

	// Abort if no negation consists of only this literal:
	if(neg_elem == 0)
		return;

	// Now loop over all conditional facts that contain this negation:
	for(cond_elem = neg_elem->get_neg()->get_occur(); cond_elem != 0;
			cond_elem = cond_elem->get_next_neg_occur()) {
		// Copy conditional fact, remove default negation:
		cf = cond_elem->get_cf();
		head_ptr = new_head;
		for(hd = cf->get_head(); hd; hd = hd->get_next_head_elem()) {
			CHECK(head_ptr >= new_head,
				"cf_c::pos_reduce: head_ptr invalid");
			CHECK(head_ptr - new_head < VER_MAX_HEAD_LIT,
				"cf_c::pos_reduce: too many head literals");
			*head_ptr++ = hd->get_fact();
		}
		cond_ptr = new_cond;
		for(cn = cf->get_cond(); cn; cn = cn->get_next_cond_elem()) {
			CHECK(cond_ptr >= new_cond,
				"cf_c::pos_reduce: cond_ptr invalid");
			CHECK(cond_ptr - new_cond < VER_MAX_COND_NEG,
				"cf_c::pos_reduce: too many negations");
			if(cn != cond_elem)
				*cond_ptr++ = cn->get_neg();
		}
		cf_c::store(new_head, (int)(head_ptr-new_head),
				new_cond, (int)(cond_ptr-new_cond));
	}
}

//=============================================================================
// Member Functions (Object Methods) of cf_hd_c:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

cf_hd_c::cf_hd_c(glit_t fact, cf_c *cf)
{
	// Initialize Members:
	Fact = fact;
	Cf = cf;
	NextHeadElem = 0;
	NextFactOccur = 0;
	PrevFactOccur = 0;
	
	// The magic number must be set before calling set_fact_prev below:
	CHECK_CODE(Magic = CF_HD_MAGIC);

	// Link to list of cf_hd_c objects for this fact:
	NextFactOccur = fact->get_pos_occur();
	fact->set_pos_occur(this);
	if(NextFactOccur != 0)
		NextFactOccur->set_prev_fact_occur(this);
	if(cf->is_cond_answer() == BOOL_FALSE)
		fact->inc_pos_occur();
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

cf_hd_c::~cf_hd_c(void)
{
	// When the destructor is called, the object must still be valid:
	CHECK(valid(), "cf_hd_c::Destructor: Bad object");

	// There is no point to remove the head element from its cond fact
	// because it gets deleted only when the entire cond fact is deleted.

	// But we need to remove it from the set of occurrences of its fact.
	if(PrevFactOccur == 0)
		Fact->set_pos_occur(NextFactOccur);
	else
		PrevFactOccur->set_next_fact_occur(NextFactOccur);
	if(NextFactOccur != 0)
		NextFactOccur->set_prev_fact_occur(PrevFactOccur);

	// Decrement number of occurrences of fact in non-answer cfs:
	if(Cf->is_cond_answer() == BOOL_FALSE)
		Fact->dec_pos_occur();

	// Invalidate the object:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
bool_t	cf_hd_c::valid(void) const
{
	// Check object pointer:
	CHECK(this != 0, "cf_hd_c::valid: NIL object");

	// Magic number must be correct:
	if(Magic != CF_HD_MAGIC)
		return(BOOL_FALSE);

	// The reference back to the owning conditional fact must be set:
	CHECK(Cf != 0, "cf_hd_c::valid: Head element belongs to no cond fact");

	// The fact must be set:
	CHECK(Fact != 0, "cf_hd_c::valid: Fact is NIL");

	// Ok:
	return(BOOL_TRUE);
}
#endif

//=============================================================================
// Member Functions (Object Methods) of cf_cn_c:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

cf_cn_c::cf_cn_c(gneg_t neg, cf_c *cf)
{
	// Initialize Members:
	Neg = neg;
	Cf = cf;
	NextCondElem = 0;
	NextNegOccur = 0;
	PrevNegOccur = 0;
	
	// The magic number must be set before calling set_neg_prev below:
	CHECK_CODE(Magic = CF_CN_MAGIC);

	// Link to list of cf_cn_c objects for the default negation:
	NextNegOccur = neg->get_occur();
	neg->set_occur(this);
	if(NextNegOccur != 0)
		NextNegOccur->set_prev_neg_occur(this);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

cf_cn_c::~cf_cn_c(void)
{
	// When the destructor is called, the object must still be valid:
	CHECK(valid(), "cf_cn_c::Destructor: Bad object");

	// There is no point to remove the body element from its cond fact
	// because it gets deleted only when the entire cond fact is deleted.

	// But it must be removed from the set of occurrences of the negation:
	if(PrevNegOccur == 0)
		Neg->set_occur(NextNegOccur);
	else
		PrevNegOccur->set_next_neg_occur(NextNegOccur);
	if(NextNegOccur != 0)
		NextNegOccur->set_prev_neg_occur(PrevNegOccur);

	// Invalidate the object:
	CHECK_CODE(Magic = 0);
}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
bool_t	cf_cn_c::valid(void) const
{
	// Check object pointer:
	CHECK(this != 0, "cf_cn_c::valid: NIL object");

	// Magic number must be correct:
	if(Magic != CF_CN_MAGIC)
		return(BOOL_FALSE);

	// The reference back to the owning conditional fact must be set:
	CHECK(Cf != 0,
		"cf_cn_c::valid: Condition element belongs to no cond fact");

	// The default negation must be set:
	CHECK(Neg != 0, "cf_cn_c::valid: Default negation is NIL");

	// Ok:
	return(BOOL_TRUE);
}
#endif

